package org.processmining.datapetrinets.datadiscovery;

import java.util.Map;
import java.util.Set;

import org.processmining.models.FunctionEstimator.DecisionTreeFunctionEstimator;
import org.processmining.models.FunctionEstimator.Type;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;

public class BasicDecisionTreeImpl extends AbstractDecisionRuleDiscovery {

	public enum NoLeafAction {
		TREAT_AS_TRUE, TREAT_AS_FALSE
	}

	private NoLeafAction noLeafAction = NoLeafAction.TREAT_AS_TRUE;

	public BasicDecisionTreeImpl(DecisionTreeConfig config, Iterable<ProjectedTrace> projectedLog,
			int numInstancesEstimate) {
		super(config, projectedLog, numInstancesEstimate);
	}

	public BasicDecisionTreeImpl(DecisionTreeConfig config, Iterable<ProjectedTrace> projectedLog,
			Map<String, Type> attributeType, Map<String, Set<String>> literalValues, int numInstancesEstimate) {
		super(config, projectedLog, attributeType, literalValues, numInstancesEstimate);
	}

	public DecisionRule discoverRulesForPlace(PetrinetGraph net, Place place) throws RuleDiscoveryException {

		Set<Transition> transitions = getTransitionPostset(net, place);

		DecisionTreeFunctionEstimator estimator = new DecisionTreeFunctionEstimator(attributeType, literalValues,
				transitions.toArray(), place.getLabel(), getNumInstanceEstimate());
		
		estimator.setTreatNoLeafAsFalse(noLeafAction == NoLeafAction.TREAT_AS_FALSE);

		return doDiscoverDecisionTree(place, transitions, estimator);
	}

	public NoLeafAction getNoLeafAction() {
		return noLeafAction;
	}

	public void setNoLeafAction(NoLeafAction noLeafAction) {
		this.noLeafAction = noLeafAction;
	}

}
