package org.processmining.datapetrinets.datadiscovery;

import java.util.Map;
import java.util.Set;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import org.processmining.models.FunctionEstimator.DecisionTreeBasedFunctionEstimator;
import org.processmining.models.FunctionEstimator.OverlappingEstimatorPairwiseDecisionTrees;
import org.processmining.models.FunctionEstimator.Type;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;

public class OverlappingPairwiseDecisionTreeImpl extends AbstractDecisionRuleDiscovery {

	public OverlappingPairwiseDecisionTreeImpl(DecisionTreeConfig config, Iterable<ProjectedTrace> projectedLog,
			int numInstancesEstimate) {
		super(config, projectedLog, numInstancesEstimate);
	}

	public OverlappingPairwiseDecisionTreeImpl(DecisionTreeConfig config, Iterable<ProjectedTrace> projectedLog,
			Map<String, Type> attributeType, Map<String, Set<String>> literalValues, int numInstancesEstimate) {
		super(config, projectedLog, attributeType, literalValues, numInstancesEstimate);
	}

	public DecisionRule discoverRulesForPlace(PetrinetGraph net, Place place) throws RuleDiscoveryException {

		Set<Transition> transitions = getTransitionPostset(net, place);

		ThreadPoolExecutor pool = new ThreadPoolExecutor(Runtime.getRuntime().availableProcessors() / 2, Runtime
				.getRuntime().availableProcessors() / 2, 60, TimeUnit.SECONDS, new LinkedBlockingQueue<Runnable>());

		try {
			DecisionTreeBasedFunctionEstimator estimator = new OverlappingEstimatorPairwiseDecisionTrees(attributeType,
					literalValues, transitions.toArray(), getNumInstanceEstimate(), place.getLabel(), pool);
			return doDiscoverDecisionTree(place, transitions, estimator);
		} finally {
			pool.shutdown();
		}
	}

}
