package org.processmining.models.FunctionEstimator;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.datapetrinets.utils.WekaTreeClassificationAdapter;
import org.processmining.datapetrinets.utils.WekaTreeClassificationAdapter.WekaCondition;
import org.processmining.datapetrinets.utils.WekaTreeClassificationAdapter.WekaLeafNode;

import weka.classifiers.AbstractClassifier;
import weka.classifiers.trees.J48;
import weka.core.Attribute;
import weka.core.Instance;
import weka.core.Instances;

@Deprecated
public class OverlappingEstimatorGlobalDecisionTree extends AbstractFunctionEstimator implements DecisionTreeBasedFunctionEstimator {

	public static class LeafProcessor {

		private final Object predictedClassObject;
		private final Integer predictedClassIndex;
		private final List<Instance> instancesAtLeaf;

		public LeafProcessor(WekaLeafNode node, DecisionTreeFunctionEstimator normalEstimator) {
			this.predictedClassObject = normalEstimator.classMapping.get(node.getClassName());
			this.predictedClassIndex = normalEstimator.classIndexMap.get(predictedClassObject);
			instancesAtLeaf = new ArrayList<>((int) node.getInstanceCount());

			for (Instance instance : normalEstimator.instances) {
				boolean fulfills = true;
				for (WekaCondition c : node.getConditions()) {
					if (!fulfillsCondition(instance, c)) {
						fulfills = false;
						break;
					}
				}
				if (fulfills) {
					instancesAtLeaf.add(instance);
				}
			}
		}

		private static boolean fulfillsCondition(Instance instance, WekaCondition condition) {
			Instances dataset = instance.dataset();
			Attribute attribute = dataset.attribute(condition.getAttributeName());
			double value = instance.value(attribute.index());
			double instanceValue;
			if (!instance.isMissing(attribute)) {
				if (attribute.isNominal()) {
					instanceValue = attribute.indexOfValue(condition.getValue());
				} else {
					instanceValue = Double.valueOf(condition.getValue());
				}

				switch (condition.getOperator()) {
					case "=" :
						return Double.compare(value, instanceValue) == 0;
					case "!=" :
						return Double.compare(value, instanceValue) != 0;
					case "<" :
						return Double.compare(value, instanceValue) < 0;
					case ">" :
						return Double.compare(value, instanceValue) > 0;
					case "<=" :
						return Double.compare(value, instanceValue) <= 0;
					case ">=" :
						return Double.compare(value, instanceValue) >= 0;
				}

				throw new IllegalArgumentException("Operator unkown in: " + condition);

			} else {
				return false;
			}
		}

		public List<Instance> getInstancesAtLeaf() {
			return instancesAtLeaf;
		}

		public Integer getPredictedClassIndex() {
			return predictedClassIndex;
		}

	}

	private final Object[] outputClasses;
	private final String decisionPointName;
	private final DecisionTreeFunctionEstimator normalEstimator;

	private Map<Object, FunctionEstimation> estimation = new HashMap<Object, FunctionEstimation>();
	private Map<String, Type> attributeType;
	private Map<String, Set<String>> literalValues;

	public OverlappingEstimatorGlobalDecisionTree(Map<String, Type> attributeType,
			Map<String, Set<String>> literalValues, Object[] outputClasses, int capacity, String name) {
		super();
		this.attributeType = attributeType;
		this.outputClasses = outputClasses;
		this.literalValues = literalValues;
		this.decisionPointName = name;
		this.normalEstimator = new DecisionTreeFunctionEstimator(attributeType, literalValues, outputClasses, name,
				capacity);
	}

	public void addInstance(Map<String, Object> variableAssignment, Object outputValue, float weight) throws Exception {
		normalEstimator.addInstance(variableAssignment, outputValue, weight);
	}
	
	public void saveInstances(File file) throws IOException {
		normalEstimator.saveInstances(file);
	}

	public Map<Object, FunctionEstimation> getFunctionEstimation(Object[] option) throws Exception {

		List<WekaLeafNode> manipulatedLeaves = new ArrayList<WekaTreeClassificationAdapter.WekaLeafNode>();

		normalEstimator.getEstimation(option);

		if (outputClasses.length > 1) {

			J48 tree = (J48) normalEstimator.getClassifier();

			WekaTreeClassificationAdapter wekaJ48Adapter = new WekaTreeClassificationAdapter(tree,
					normalEstimator.instances, attributeType);

			final DecisionTreeFunctionEstimator overlappingClassesEstimator = new DecisionTreeFunctionEstimator(
					attributeType, literalValues, outputClasses, decisionPointName, normalEstimator.getNumInstances());

			overlappingClassesEstimator.setBinarySplit(normalEstimator.binarySplit);
			overlappingClassesEstimator.setNumFolds(normalEstimator.numFoldErrorPruning);
			overlappingClassesEstimator.setMinNumObj(normalEstimator.minNumInstancePerLeaf);
			overlappingClassesEstimator.setUnpruned(normalEstimator.unpruned);
			overlappingClassesEstimator.setCrossValidate(normalEstimator.crossValidate);

			for (WekaLeafNode leafNode : wekaJ48Adapter.traverseLeafNodes()) {

				// Add original leaf node
				manipulatedLeaves.add(leafNode);

				LeafProcessor leafProcessor = new LeafProcessor(leafNode, normalEstimator);

				// Build new data set only retaining the wrongly classified instances
				for (Instance instance : leafProcessor.getInstancesAtLeaf()) {
					Integer realClassIndex = (int) instance.value(instance.classIndex());
					Object realClass = normalEstimator.classIndexMap.inverse().get(realClassIndex);
					if (realClassIndex != leafProcessor.getPredictedClassIndex()) {
						overlappingClassesEstimator.addWekaInstance(instance, realClass, 1.0f);
					}
				}
			}

			AbstractClassifier overlappingTree = overlappingClassesEstimator.createClassifier(null, false);
			WekaTreeClassificationAdapter overlappingTreeAdapter = new WekaTreeClassificationAdapter(
					(J48) overlappingTree, overlappingClassesEstimator.instances, attributeType);

			if (overlappingTreeAdapter.treeDepth() > 0) {
				// Add newly found overlapping conditions
				for (WekaLeafNode newLeaf : overlappingTreeAdapter.traverseLeafNodes()) {
					manipulatedLeaves.add(newLeaf);
					System.out.println("New leaf: " + newLeaf);
				}
			}
		}

		estimation.clear();
		for (WekaLeafNode leaf : manipulatedLeaves) {
			Object leafClass = normalEstimator.classMapping.get(leaf.getClassName());
			FunctionEstimation existingCondition = estimation.get(leafClass);
			if (existingCondition != null) {
				GuardExpression newCondition = GuardExpression.Operation.or(existingCondition.getExpression(), leaf.getExpression());
				estimation.put(leafClass, FunctionEstimation.Factory.create(newCondition, -1d));
			} else {
				GuardExpression expression = leaf.getExpression();
				if (expression != null) {
					estimation.put(leafClass, FunctionEstimation.Factory.create(expression, -1d));
				}
			}
		}
		return estimation;
	}

	public String toString() {
		return normalEstimator.tree.toString();
	}

	public double computeFMeasure() {
		return computeQualityMeasure();
	}
	
	public double computeQualityMeasure() {
		return -1;
	}

	public String getQualityMeasureName() {
		return "No Quality Measure Defined";
	}

	public int getNumInstances() {
		return normalEstimator.getNumInstances();
	}

	public void setMinNumObj(int minNumObj) {
		normalEstimator.setMinNumObj(minNumObj);
	}

	public void setUnpruned(boolean unpruned) {
		normalEstimator.setUnpruned(unpruned);
	}

	public void setBinarySplit(boolean selected) {
		normalEstimator.setBinarySplit(selected);
	}

	public void setCrossValidate(boolean validate) {
		normalEstimator.setCrossValidate(validate);
	}

	public void setConfidenceFactor(float confidenceThreshold) {
		normalEstimator.setConfidenceFactor(confidenceThreshold);
	}

	public void setNumFolds(int numFoldErrorPruning) {
		normalEstimator.setNumFolds(numFoldErrorPruning);
	}

}
