package org.processmining.models.FunctionEstimator;

import java.io.File;
import java.io.IOException;
import java.util.ArrayDeque;
import java.util.Comparator;
import java.util.Deque;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.datapetrinets.utils.WekaTreeClassificationAdapter;
import org.processmining.datapetrinets.utils.WekaTreeClassificationAdapter.WekaCondition;
import org.processmining.datapetrinets.utils.WekaTreeClassificationAdapter.WekaLeafNode;

import com.google.common.collect.BiMap;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;

import weka.classifiers.Evaluation;
import weka.classifiers.trees.J48;
import weka.core.Attribute;
import weka.core.Instance;
import weka.core.Instances;

@Deprecated
public class OverlappingEstimatorSimpleLabelExtension extends AbstractFunctionEstimator implements
		DecisionTreeBasedFunctionEstimator {

	private final Object[] outputClasses;
	private final String decisionPointName;
	private final DecisionTreeFunctionEstimator normalEstimator;

	private Map<Object, FunctionEstimation> estimation;
	private Map<String, Type> attributeType;

	public OverlappingEstimatorSimpleLabelExtension(Map<String, Type> attributeType,
			Map<String, Set<String>> literalValues, Object[] outputClasses, int capacity, String name) {
		super();
		this.attributeType = attributeType;
		this.outputClasses = outputClasses;
		decisionPointName = name;
		normalEstimator = new DecisionTreeFunctionEstimator(attributeType, literalValues, outputClasses, name, capacity);
	}

	public void addInstance(Map<String, Object> variableAssignment, Object outputValue, float weight) throws Exception {
		normalEstimator.addInstance(variableAssignment, outputValue, weight);
	}

	public void saveInstances(File file) throws IOException {
		normalEstimator.saveInstances(file);
	}

	public Map<Object, FunctionEstimation> getFunctionEstimation(Object[] option) throws Exception {

		estimation = normalEstimator.getFunctionEstimation(option);

		if (outputClasses.length > 1) {

			J48 tree = (J48) normalEstimator.getClassifier();
			Set<Object> classSet = ImmutableSet.copyOf(outputClasses);

			WekaTreeClassificationAdapter wekaJ48Adapter = new WekaTreeClassificationAdapter(tree,
					normalEstimator.instances, attributeType);

			if (wekaJ48Adapter.treeDepth() > 0) {

				ImmutableList<WekaLeafNode> leavesSortedByWrongInstances = wekaJ48Adapter.traverseLeafNodes()
						.toSortedList(new Comparator<WekaLeafNode>() {

							public int compare(WekaLeafNode o1, WekaLeafNode o2) {
								return Double.compare(o2.getWrongInstanceCount(), o1.getWrongInstanceCount());
							}
						});

				Deque<WekaLeafNode> leavesExtended = new ArrayDeque<>();

				for (Object clazz : classSet) {

					FunctionEstimation oldEstimation = estimation.get(clazz);
					double currentFScore = oldEstimation != null ? oldEstimation.getQualityMeasure() : 0.0d;
					GuardExpression currentGuard = oldEstimation != null ? oldEstimation.getExpression() : null;

					// Try to extend all leaves with this class
					for (WekaLeafNode node : leavesSortedByWrongInstances) {

						if (normalEstimator.classMapping.get(node.getClassName()) == clazz) {
							// Do not extend leaf pointing to the same class
							continue;
						}

						leavesExtended.push(node);

						// Measure the change of the f-score for the current transition 
						double newFScore = reComputeScore(clazz, leavesExtended, tree, normalEstimator.instances,
								normalEstimator.getEvaluation(), normalEstimator.classIndexMap.inverse());

						if (Double.compare(newFScore, currentFScore) > 0) {
							System.out.println(String.format(
									"Improving f-score for transition %s (%s -> %s) by extending leaf %s", clazz,
									currentFScore, newFScore, node));
							// This improves the f-score of the current class
							currentFScore = newFScore;
							if (currentGuard != null) {
								GuardExpression combinedExpression = GuardExpression.Operation.or(currentGuard,
										node.getExpression());
								estimation.put(clazz, FunctionEstimation.Factory.create(combinedExpression, newFScore));
							} else {
								estimation.put(clazz,
										FunctionEstimation.Factory.create(node.getExpression(), newFScore));
							}

						} else {
							System.out.println(String.format(
									"Cannot improve f-score (%s -> %s) for %s by extending leaf %s", currentFScore,
									newFScore, clazz, node));
							leavesExtended.pop();
						}

					}

					if (!leavesExtended.isEmpty()) {
						System.out.println(String.format("Extended %s with levaes: %s", clazz, leavesExtended));
					}

					leavesExtended.clear();

				}

				return estimation;
			} else {
				// Nothing to do, no decision tree learned
				return estimation;
			}

		} else {
			return estimation;
		}
	}

	private double reComputeScore(Object clazz, Deque<WekaLeafNode> leavesExtended, J48 tree, Instances instances,
			Evaluation oldEvaluation, BiMap<Integer, Object> indexToClass) throws Exception {

		int falsePositives = (int) oldEvaluation.numFalsePositives(indexToClass.inverse().get(clazz));
		int falseNegatives = (int) oldEvaluation.numFalseNegatives(indexToClass.inverse().get(clazz));
		int truePositives = (int) oldEvaluation.numTruePositives(indexToClass.inverse().get(clazz));

		for (Instance instance : instances) {
			boolean isClassified = conditionsTrue(leavesExtended, instance);
			boolean sameClass = isClass(clazz, instance, indexToClass);
			if (isClassified && sameClass) {
				// This means that an instance is now a TP instead of a FN (Recall improves)
				truePositives++;
				falseNegatives--;
			} else if (isClassified && !sameClass) {
				// This means we added a FP that we did not have before (Precision decreases)
				// This could not have been a false positive before as the initial decision tree is mutually exclusive! 
				falsePositives++;
			}
		}

		double differentialFScore = calculateFScore(truePositives, falsePositives, falseNegatives);
		double fullFScore = reCalculateFullScore(clazz, leavesExtended, tree, instances, indexToClass);
		assert fullFScore == differentialFScore : "Differential F-Score is wrong " + differentialFScore
				+ " full score " + fullFScore;

		return differentialFScore;
	}

	private double reCalculateFullScore(Object clazz, Deque<WekaLeafNode> leavesExtended, J48 tree,
			Instances instances, BiMap<Integer, Object> indexToClass) throws Exception {
		int falsePositives = 0;
		int falseNegatives = 0;
		int truePositives = 0;

		for (Instance instance : instances) {

			Object originalClassification = indexToClass.get((int) tree.classifyInstance(instance));

			boolean originalClassifiedAsClazz = originalClassification == clazz;
			boolean newClassifiedAsClazz = conditionsTrue(leavesExtended, instance);

			boolean isClassifiedAsClazz = newClassifiedAsClazz || originalClassifiedAsClazz;

			assert (originalClassifiedAsClazz == true ? newClassifiedAsClazz == false : true) : "Old classifier "
					+ tree + " classifies  " + instance + " as " + (originalClassifiedAsClazz ? clazz : "not" + clazz)
					+ ", so new extended leaf should not classify it as such " + leavesExtended;

			boolean wasClazz = isClass(clazz, instance, indexToClass);

			if (isClassifiedAsClazz && wasClazz) {
				truePositives++;
			} else if (isClassifiedAsClazz && !wasClazz) {
				falsePositives++;
			} else if (!isClassifiedAsClazz && wasClazz) {
				falseNegatives++;
			}
		}

		return calculateFScore(truePositives, falsePositives, falseNegatives);
	}

	private static double calculateFScore(int tp, int fp, int fn) {
		double precision;
		double recall;

		if (fp == 0.0d) {
			precision = 1.0d;
		} else {
			precision = ((double) tp) / (tp + fp);
		}

		if (fn == 0.0d) {
			recall = 1.0d;
		} else {
			recall = ((double) tp) / (tp + fn);
		}
		if (precision == 0.0f && recall == 0.0f) {
			return 0.0f;
		} else {
			double fscore = (2 * precision * recall) / (precision + recall);
			return fscore;
		}
	}

	private boolean isClass(Object clazz, Instance instance, Map<Integer, Object> indexToClass) {
		int priorWeka = (int) instance.value(instance.classIndex());
		Object priorClass = indexToClass.get(priorWeka);
		return priorClass == clazz;
	}

	private boolean conditionsTrue(Deque<WekaLeafNode> leavesExtended, Instance instance) {
		for (WekaLeafNode l : leavesExtended) {
			for (WekaCondition c : l.getConditions()) {
				if (!fulfillsCondition(instance, c)) {
					return false;
				}
			}
		}
		return true;
	}

	private boolean fulfillsCondition(Instance instance, WekaCondition condition) {
		Instances dataset = instance.dataset();
		Attribute attribute = dataset.attribute(condition.getAttributeName());
		double value = instance.value(attribute.index());
		double instanceValue;
		if (!instance.isMissing(attribute)) {
			if (attribute.isNominal()) {
				instanceValue = attribute.indexOfValue(condition.getValue());
			} else {
				instanceValue = Double.valueOf(condition.getValue());
			}

			switch (condition.getOperator()) {
				case "=" :
					return Double.compare(value, instanceValue) == 0;
				case "!=" :
					return Double.compare(value, instanceValue) != 0;
				case "<" :
					return Double.compare(value, instanceValue) < 0;
				case ">" :
					return Double.compare(value, instanceValue) > 0;
				case "<=" :
					return Double.compare(value, instanceValue) <= 0;
				case ">=" :
					return Double.compare(value, instanceValue) >= 0;
			}

			throw new IllegalArgumentException("Operator unkown in: " + condition);

		} else {
			return false;
		}
	}

	public String toString() {
		return normalEstimator.tree.toString();
	}

	public double computeFMeasure() {
		return computeQualityMeasure();
	}

	public double computeQualityMeasure() {
		double sumFMeasures = 0;
		double toConsider = 0;
		for (Entry<Object, FunctionEstimation> e : estimation.entrySet()) {
			if (e.getValue().getQualityMeasure() > 0) {
				sumFMeasures += 1 / e.getValue().getQualityMeasure();
				toConsider++;
			}
		}
		if (toConsider != 0)
			return toConsider / sumFMeasures;
		else
			return 0;
	}

	public String getQualityMeasureName() {
		return "F-Score";
	}

	public int getNumInstances() {
		return normalEstimator.getNumInstances();
	}

	public void setMinNumObj(int i) {
		normalEstimator.setMinNumObj(i);
	}

	public void setUnpruned(boolean b) {
		normalEstimator.setUnpruned(b);
	}

	public void setBinarySplit(boolean selected) {
		normalEstimator.setBinarySplit(selected);
	}

	public void setCrossValidate(boolean validate) {
		normalEstimator.setCrossValidate(validate);
	}

	public void setConfidenceFactor(float confidenceThreshold) {
		normalEstimator.setConfidenceFactor(confidenceThreshold);
	}

	public void setNumFolds(int numFoldErrorPruning) {
		normalEstimator.setNumFolds(numFoldErrorPruning);
	}

}
