package org.processmining.plugins.petrinetwithdata.decisionminer;

import java.util.Map;

import org.deckfour.xes.model.XTrace;
import org.processmining.framework.plugin.Progress;
import org.processmining.models.FunctionEstimator.FunctionEstimator;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.replayer.replayresult.AllSyncReplayResult;

import com.google.common.util.concurrent.AtomicLongMap;

/**
 * 
 * @author Mirko Polato
 *
 */
public class WeightedTraceProcessor extends AbstractTraceProcessor {

	private AllSyncReplayResult alignments;
	private boolean allAlignments = false;

	public WeightedTraceProcessor(PetrinetGraph net, 
								  XTrace xTrace, 
								  Map<Place, FunctionEstimator> estimators,
								  AllSyncReplayResult alignments, 
								  AtomicLongMap<Transition> numberOfExecutions, 
								  Map<Transition, AtomicLongMap<String>> numberOfWritesPerTransition, 
								  Progress progress, 
								  float fitness) {
		
		super(net, xTrace, estimators, numberOfExecutions, numberOfWritesPerTransition, progress);
		
		this.alignments = alignments;
		this.fitness = fitness;
	}
	
	public WeightedTraceProcessor(PetrinetGraph net, 
			  XTrace xTrace, 
			  Map<Place, FunctionEstimator> estimators,
			  AllSyncReplayResult alignments, 
			  AtomicLongMap<Transition> numberOfExecutions, 
			  Map<Transition, AtomicLongMap<String>> numberOfWritesPerTransition,  
			  Progress progress, 
			  float fitness,
			  boolean allAlignments) {

		super(net, xTrace, estimators, numberOfExecutions, numberOfWritesPerTransition, progress);
		
		this.alignments = alignments;
		
		if (allAlignments) {
			this.weight = 1.0f / alignments.getNodeInstanceLst().size();
			this.allAlignments = true;
		}
		
		//this.weight = this.fitness / alignments.getNodeInstanceLst().size();
		this.fitness = fitness;
	}


	public void run() {
		
		if (allAlignments) {
			
			for (int i = 0; i < alignments.getNodeInstanceLst().size(); ++i) {
				processAlignment(alignments.getStepTypesLst().get(i), alignments.getNodeInstanceLst().get(i));
			}
			
		} else {
			
			int i = Math.min((int)(Math.random() * alignments.getNodeInstanceLst().size()), alignments.getNodeInstanceLst().size() - 1);
			//System.out.println("Debug - MAX: " + alignments.getNodeInstanceLst().size());
			//System.out.println("Debug - RANDOM: " + i);
			
			processAlignment(alignments.getStepTypesLst().get(i), alignments.getNodeInstanceLst().get(i));
		}
	}
	
}
